/*
  HD44780 LCD Driver

  2009-01-20

  T. Nakagawa
*/


#include <avr/io.h>
#include <util/delay.h>
#include "lcd.h"


#define RS_PORT PORTB
#define RS_PIN PB4
#define E_PORT PORTB
#define E_PIN PB5
#define DATA_PORT PORTB
#define DATA_PIN PB0


static void lcd_setadrs(unsigned char adrs);
static void hd44780_outcmd(unsigned char d);
static void hd44780_outdat(unsigned char d);


void lcd_init(void) {
  _delay_ms(15);
  hd44780_outcmd(0x03);
  _delay_ms(4.1);
  hd44780_outcmd(0x03);
  _delay_ms(0.1);
  hd44780_outcmd(0x03);

  /* 4bit */
  hd44780_outcmd(0x02);
  _delay_us(40);
  /* 4bit, 2lines, 5x7dots */
  hd44780_outcmd(0x02);
  hd44780_outcmd(0x08);
  _delay_us(40);
  /* display-on, cursor-off, blink-off */
  hd44780_outcmd(0x00);
  hd44780_outcmd(0x0c);
  _delay_us(40);
  /* Clear */
  hd44780_outcmd(0x00);
  hd44780_outcmd(0x01);
  _delay_ms(1.64);
  /* cursor-increment, without-display-shift */
  hd44780_outcmd(0x00);
  hd44780_outcmd(0x06);
  _delay_us(40);

  return;
}


void lcd_move(unsigned char p) {
  hd44780_outcmd(0x08 | (p >> 4));
  hd44780_outcmd(p & 0x0f);
  _delay_us(40);

  return;
}


void lcd_putchar(unsigned char c) {
  hd44780_outdat(c >> 4);
  hd44780_outdat(c & 0x0f);
  _delay_us(40);

  return;
}


void lcd_setpat(unsigned char adrs, unsigned char pat0, unsigned char pat1, unsigned char pat2, unsigned char pat3, unsigned char pat4, unsigned char pat5, unsigned char pat6, unsigned char pat7) {

  lcd_setadrs((adrs << 3) | 0x00);
  lcd_putchar(pat0);
  lcd_setadrs((adrs << 3) | 0x01);
  lcd_putchar(pat1);
  lcd_setadrs((adrs << 3) | 0x02);
  lcd_putchar(pat2);
  lcd_setadrs((adrs << 3) | 0x03);
  lcd_putchar(pat3);
  lcd_setadrs((adrs << 3) | 0x04);
  lcd_putchar(pat4);
  lcd_setadrs((adrs << 3) | 0x05);
  lcd_putchar(pat5);
  lcd_setadrs((adrs << 3) | 0x06);
  lcd_putchar(pat6);
  lcd_setadrs((adrs << 3) | 0x07);
  lcd_putchar(pat7);

  lcd_move(0);

  return;
}


static void lcd_setadrs(unsigned char adrs) {
  hd44780_outcmd(0x04 | (adrs >> 4));
  hd44780_outcmd(adrs & 0x0f);
  _delay_us(40);

  return;
}


static void hd44780_outcmd(unsigned char d) {
  RS_PORT &= ~_BV(RS_PIN);
  _delay_us(0.04);
  E_PORT |= _BV(E_PIN);
  DATA_PORT = ((DATA_PORT & (~0x0f << DATA_PIN)) | (d << DATA_PIN));
  _delay_us(0.22);
  E_PORT &= ~_BV(E_PIN);
  _delay_us(0.28);

  return;
}


static void hd44780_outdat(unsigned char d) {
  RS_PORT |= _BV(RS_PIN);
  _delay_us(0.04);
  E_PORT |= _BV(E_PIN);
  DATA_PORT = ((DATA_PORT & (~0x0f << DATA_PIN)) | (d << DATA_PIN));
  _delay_us(0.22);
  E_PORT &= ~_BV(E_PIN);
  _delay_us(0.28);

  return;
}
