/*
  Lux Meter

  T. Nakagawa

  2009-01-19
*/


#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "suart.h"
#include "lcd.h"


#define AREF 2.811


static unsigned short sense(void);
static void put_bar(unsigned short val);
static void put_digit(unsigned short val);


prog_uint16_t logtbl[48] = {
  1, 2, 2, 3, 3, 4, 5, 6, 8, 10, 13, 16, 20, 25, 32, 40, 51, 64, 81, 
  102, 128, 161, 203, 256, 323, 406, 512, 645, 813, 1024, 1290, 1625, 
  2048, 2580, 3251, 4096, 5161, 6502, 8192, 10321, 13004, 16384, 20642, 
  26008, 32768, 41284, 52015, 65535
};


int main(void) {
  unsigned char i;
  unsigned short val;

  /* |[g */
  DDRA = 0x04;
  DDRB = 0xff;

  /* LCD */
  lcd_init();
  lcd_setpat(0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00);
  lcd_setpat(0x01, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10);
  lcd_setpat(0x02, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14);
  lcd_setpat(0x03, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15);

  /* ADC */
  ADMUX = 0xcc;
  ADCSR = 0x83;

  for (i = 0; ; i = (i + 1) & 0x03) {
    /*  */
    val = sense();

    /* o[Ot\ */
    put_bar(val);

    if (i != 3) {
      _delay_ms(10);
      continue;
    }

    /* l\ & UARTM */
    put_digit(val);
  }

  return 0;
}


static unsigned short sense(void) {
  unsigned char i, m = 0;
  unsigned short v, sum, val;

  sum = 0;
  for (i = 0; i < 64; i++) {
    /* Wݒ */
    if (m == 0) {	/* 100kohm (85.33lux) */
      DDRA = ((DDRA & 0x0f) | 0x80);
    } else if (m == 1) {	/* 10kohm (853.3lux) */
      DDRA = ((DDRA & 0x0f) | 0x40);
    } else if (m == 2) {	/* 1kohm (8533lux) */
      DDRA = ((DDRA & 0x0f) | 0x20);
    } else if (m == 3) {	/* 100ohm (85330lux) */
      DDRA = ((DDRA & 0x0f) | 0x10);
    }
    _delay_ms(1);

    /* ADϊ */
    ADCSR |= _BV(ADSC);
    while (ADCSR & _BV(ADSC)) ;
    v = ADCL;
    v |= ADCH << 8;

    /* W؂ւ */
    if (m != 3 && v == 1023) {
      m++;
      i = -1;
      sum = 0;
      continue;
    }

    sum += v;
  }

  v = (unsigned short)((unsigned long)sum * (unsigned long)(AREF * 1e4 + 0.5) / (3UL * 1024 * 64));
  val = ((v << 2) | m);

  return val;
}


static void put_bar(unsigned short val) {
  unsigned int i;
  unsigned int len;
  unsigned char r;

  r = val & 0x0003;
  val >>= 2;
  /* 0-100,000lux0-65535ɐK */
  val = (unsigned short)((unsigned long)val * 65535UL / 10000UL);
  if (r == 0) {
    val /= 1000;
  } else if (r == 1) {
    val /= 100;
  } else if (r == 2) {
    val /= 10;
  }

  len = 0;
  for (i = 0; i < 48; i++) {
    if (val < pgm_read_word(logtbl + i)) break;
    len++;
  }

  lcd_move(0x40);
  for (i = 0; i < len / 3; i++) {
    lcd_putchar(3);
  }
  lcd_putchar(len % 3);
  for (; i < 16; i++) {
    lcd_putchar(' ');
  }

  return;
}


static void put_digit(unsigned short val) {
  unsigned char v0, v1, v2, v3;
  unsigned char r;

  r = val & 0x0003;
  val >>= 2;

  v0 = val % 10 + '0';
  val /= 10;
  v1 = val % 10 + '0';
  val /= 10;
  v2 = val % 10 + '0';
  val /= 10;
  v3 = val + '0';

  if (v3 == '0') {
    v3 = ' ';
    if (v2 == '0') {
      v2 = ' ';
      if (v1 == '0') {
        v1 = ' ';
      }
    }
  }

  lcd_move(0);
  lcd_putchar(v3);
  suart_putchar(v3);
  lcd_putchar(v2);
  suart_putchar(v2);
  if (r == 0) {
    lcd_putchar('.');
    suart_putchar('.');
  }
  lcd_putchar(v1);
  suart_putchar(v1);
  if (r == 1) {
    lcd_putchar('.');
    suart_putchar('.');
  }
  lcd_putchar(v0);
  suart_putchar(v0);
  if (r == 2) {
    lcd_putchar('.');
    suart_putchar('.');
  }
  lcd_putchar('0');
  suart_putchar('0');
  if (r == 3) {
    lcd_putchar('.');
    suart_putchar('.');
  }
  lcd_putchar(' ');
  lcd_putchar('L');
  lcd_putchar('U');
  lcd_putchar('X');
  suart_putchar('\r');
  suart_putchar('\n');

  return;
}
